unit main;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Jpeg, StdCtrls, ExtCtrls, ComCtrls, ShellApi, Math, rkView, ImgList,
  AppEvnts;

type
  PCacheItem = ^TCacheItem;
  TCacheItem = record
    Idx: Integer;
    Size: Integer;
    Age: TDateTime;
    Scale: Integer;
    Bmp: TBitmap;
  end;

  PThumbData = ^TThumbData;
  TThumbData = record
    Name: string;
    PathID: SmallInt;
    ThumbWidth: Word;
    ThumbHeight: Word;
    Rating: Byte;
    Size: Integer;
    Modified: TDateTime;
    IWidth, IHeight: Word;
    Image: TObject;
  end;

  TfrmMain = class(TForm)
    stbMain: TStatusBar;
    pnlTop: TPanel;
    pnlBottom: TPanel;
    viewMain: TrkView;
    proMain: TProgressBar;
    labInfo: TLabel;
    labText: TLabel;
    btnAdd: TButton;
    tbSize: TTrackBar;
    imgTop: TImage;
    imgBottom: TImage;
    editView: TEdit;
    ilStars: TImageList;
    chkRating: TCheckBox;
    appMain: TApplicationEvents;
    procedure UpdateStatus;
    procedure BiResample(Src, Dest: TBitmap; Sharpen: Boolean);
    function  ThumbImage(Filename: string): TBitmap;
    procedure AddImages(Files: TStringlist);
    procedure btnAddClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure ItemPaintBasic(Canvas: TCanvas; R: TRect; State: TsvItemState);
    procedure GenCellColors;
    procedure viewMainCellPaint(Sender: TObject; Canvas: TCanvas; Cell: TRect;
      IdxA, Idx: Integer; State: TsvItemState);
    procedure viewMainSelecting(Sender: TObject; Count: Integer);
    procedure tbSizeChange(Sender: TObject);
    procedure viewMainMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure viewMainResize(Sender: TObject);
    procedure viewMainMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure viewMainKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormResize(Sender: TObject);
    procedure chkRatingClick(Sender: TObject);
    procedure viewMainCellHit(Sender: TObject; Canvas: TCanvas; Idx, x,
      y: Integer; var Selected: Boolean);
    procedure editViewKeyPress(Sender: TObject; var Key: Char);
    procedure editViewKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure editViewExit(Sender: TObject);
    procedure appMainDeactivate(Sender: TObject);
  private
    { Private declarations }
    ThumbSizeW, ThumbSizeH: Integer;
    procedure SetThumbSize(Value: Integer; UpdateTrackbar: Boolean);
    function ThumbBmp(idx: Integer): TBitmap;
    procedure PurgeThumbs;
    procedure ShowStat;
    procedure ClearThumbsPool;
  public
    { Public declarations }
    Items: TList;
    Paths: TStringList;
    // Colors
    cGSelectedStart,
    cGSelectedEnd,
    cGHotStart,
    cGHotEnd,
    cGDisabledStart,
    cGDisabledEnd,
    cGHeaderStart,
    cGHeaderEnd,
    cGHeaderHotStart,
    cGHeaderHotEnd,
    cGHeaderSelStart,
    cGHeaderSelEnd,
    cHot,
    cSelected,
    cDisabled,
    cBackground,
    cLineHighLight: TColor;
    cShadeSelect: TColor;
    cShadeDisabled: TColor;
    CellShade0: TColor;
    CellShade1: TColor;
    CellShade2: TColor;
    CellShade3: TColor;
    CellShade4: TColor;
    CellBkgColor: TColor;
    CellBrdColor: array[Boolean, Boolean] of TColor;
    procedure VistaGradient(const Img: TImage; const c1, c2: Byte;
      const c3, c4: TColor; const Center, Reverse: Boolean);
  end;

var
  frmMain: TfrmMain;
  Files: TStringList;
  CellJpeg: TJpegImage;
  WI, HI, TW, TH, HSX, vIdx: Integer;
  CellScale: Integer;
  CellStyle: Integer;
  CellEditIdx: Integer;
  EditId, EditIdx: Integer;
  ThumbsPool: TList;
  PoolSize, MaxPool: Integer;
  HotStars: Integer;

implementation

uses
  BrowseForFolderU, rmkGradient, formStat;

{$R *.dfm}


function BytesToStr(const i64Size: Int64): string;
const
  i64GB = 1024 * 1024 * 1024;
  i64MB = 1024 * 1024;
  i64KB = 1024;
begin
  if i64Size div i64GB > 0 then Result := Format('%.1f GB', [i64Size / i64GB])
  else if i64Size div i64MB > 0 then Result := Format('%.2f MB', [i64Size / i64MB])
  else if i64Size div i64KB > 0 then Result := Format('%.0f kB', [i64Size / i64KB])
  else Result := IntToStr(i64Size) + ' byte';
end;

procedure TfrmMain.VistaGradient(const Img: TImage; const c1, c2: Byte;
  const c3, c4: TColor; const Center, Reverse: Boolean);
type
  PRGB = ^TRGB;
  TRGB = record b, g, r : Byte; end;
  PRGBArray = ^TRGBArray;
  TRGBARRAY = array[0..0] of TRGB;
var
  rc1, gc1, bc1, rc2, gc2, bc2, rc3, gc3, bc3: Integer;
  x, y, w, h, j: Integer;
  i, w1: Integer;
  Row: PRGBArray;
  C: TRGB;
  slMain, slSize, slPtr: Integer;
  Color: Integer;
  QCol: Array of TRGB;
  g: Byte;
  bmp: TBitmap;
begin
  bmp:= TBitmap.Create;
  bmp.PixelFormat:= pf24Bit;
  bmp.Width := Img.Width;
  bmp.Height := Img.Height;
  h:= bmp.Height;
  w:= bmp.Width;
  Color:= ColorToRGB(c3);
  rc1 := Byte(Color);
  gc1 := Byte(Color shr 8);
  bc1 := Byte(Color shr 16);
  Color:= ColorToRGB(c4);
  rc2 := Byte(Color);
  gc2 := Byte(Color shr 8);
  bc2 := Byte(Color shr 16);
  SetLength(QCol, h);
  for i := 0 to h - 1 do begin
    if Reverse then begin
      QCol[i].r:= Byte(rc1 + (((rc2 - rc1) * i) div h));
      QCol[i].g:= Byte(gc1 + (((gc2 - gc1) * i) div h));
      QCol[i].b:= Byte(bc1 + (((bc2 - bc1) * i) div h));
    end else begin
      QCol[i].r:= Byte(rc2 + (((rc1 - rc2) * i) div h));
      QCol[i].g:= Byte(gc2 + (((gc1 - gc2) * i) div h));
      QCol[i].b:= Byte(bc2 + (((bc1 - bc2) * i) div h));
    end;
  end;
  w1:= w - 1;
  if Center then w:= (w shr 1) + (w and 1);
  slMain:= Integer(bmp.ScanLine[0]);             // Init scanline accsess
  slSize:= Integer(bmp.ScanLine[1]) - slMain;
  for x:= 0 to w - 1 do begin                    // Paint gradient
    j:= (255 - c1) + MulDiv(c1, x, w);
    slPtr:= slMain;
    for y := 0 to h - 1 do begin
      Row := PRGBArray(slPtr);
      Row[x].r:= j * (QCol[y].r - rc1) shr 8 + rc1;
      Row[x].g:= j * (QCol[y].g - gc1) shr 8 + gc1;
      Row[x].b:= j * (QCol[y].b - bc1) shr 8 + bc1;
      if (Center) and (x < (w1 - x)) then begin
        Row[w1 - x].r:= Row[x].r;
        Row[w1 - x].g:= Row[x].g;
        Row[w1 - x].b:= Row[x].b;
      end;
      slPtr:= slPtr + slSize;
    end;
  end;
  QCol:= nil;
  Img.Picture.Bitmap.Assign(bmp);
  bmp.Free;
end;

procedure TfrmMain.UpdateStatus;
var
  i: integer;
  n: Int64;
  item: PThumbData;
begin
  n:= 0;
  for i:= 0 to viewMain.Selection.Count - 1 do begin
    item:= Items[viewMain.Selection[i]];
    n:= n + item.Size;
  end;
  labInfo.Caption:= IntToStr(viewMain.Items.Count) + ' items, ' +
    IntToStr(viewMain.Selection.Count) + ' selected (' + BytesToStr(n) + ')';
end;

// Recursive procedure to build a list of files
procedure FindFiles(FilesList: TStringList; StartDir: string);
var
  SR: TSearchRec;
  DirList: TStringList;
  IsFound: Boolean;
  i: integer;
  Ex: String;
begin
  if StartDir[length(StartDir)] <> '\' then StartDir := StartDir + '\';
  IsFound := FindFirst(StartDir + '*.*', faAnyFile - faDirectory, SR) = 0;
  while IsFound do begin
    Ex:= LowerCase(ExtractFileExt(SR.Name));
    if Ex = '.jpg' then FilesList.Add(StartDir + SR.Name);
    IsFound := FindNext(SR) = 0;
  end;
  FindClose(SR);
  // Build a list of subdirectories
  DirList := TStringList.Create;
  IsFound := FindFirst(StartDir + '*.*', faAnyFile, SR) = 0;
  while IsFound do begin
    if ((SR.Attr and faDirectory) <> 0) and (SR.Name[1] <> '.') then
      DirList.Add(StartDir + SR.Name);
    IsFound := FindNext(SR) = 0;
  end;
  FindClose(SR);
  // Scan the list of subdirectories
  for i := 0 to DirList.Count - 1 do FindFiles(FilesList, DirList[i]);
  DirList.Free;
end;

// http://www.swissdelphicenter.ch/en/showcode.php?id=1698
function ReadMWord(f: TFileStream): Word;
type
  TMotorolaWord = record
    case Byte of
      0: (Value: Word);
      1: (Byte1, Byte2: Byte);
  end;
var
  MW: TMotorolaWord;
begin
  { It would probably be better to just read these two bytes in normally }
  { and then do a small ASM routine to swap them.  But we aren't talking }
  { about reading entire files, so I doubt the performance gain would be }
  { worth the trouble. }
  f.read(MW.Byte2, SizeOf(Byte));
  f.read(MW.Byte1, SizeOf(Byte));
  Result := MW.Value;
end;

procedure GetJPGSize(const sFile: string; var wWidth, wHeight: Integer);
const
  ValidSig: array[0..1] of Byte = ($FF, $D8);
  Parameterless = [$01, $D0, $D1, $D2, $D3, $D4, $D5, $D6, $D7];
var
  Sig: array[0..1] of byte;
  f: TFileStream;
  x: integer;
  Seg: byte;
  Dummy: array[0..15] of byte;
  Len: word;
  ReadLen: LongInt;
begin
  FillChar(Sig, SizeOf(Sig), #0);
  f := TFileStream.Create(sFile, fmOpenRead);
  try
    ReadLen := f.read(Sig[0], SizeOf(Sig));
    for x := Low(Sig) to High(Sig) do
      if Sig[x] <> ValidSig[x] then ReadLen := 0;
    if ReadLen > 0 then
    begin
      ReadLen := f.read(Seg, 1);
      while (Seg = $FF) and (ReadLen > 0) do
      begin
        ReadLen := f.read(Seg, 1);
        if Seg <> $FF then
        begin
          if (Seg = $C0) or (Seg = $C1) then
          begin
            ReadLen := f.read(Dummy[0], 3); { don't need these bytes }
            wHeight := ReadMWord(f);
            wWidth  := ReadMWord(f);
          end
          else
          begin
            if not (Seg in Parameterless) then
            begin
              Len := ReadMWord(f);
              f.Seek(Len - 2, 1);
              f.read(Seg, 1);
            end
            else
              Seg := $FF; { Fake it to keep looping. }
          end;
        end;
      end;
    end;
  finally
    f.Free;
  end;
end;

function CalcThumbSize(w, h, tw, th: integer): TPoint;
begin
  Result.X := 0;
  Result.Y := 0;
  if (w < tw) and (h < th) then begin
    Result.X := w;
    Result.Y := h;
  end
  else if (w = 0) or (h = 0) then Exit
  else  begin
    if w > h then begin
      if w < tw then tw := w;
      Result.X := tw;
      Result.Y := Trunc(tw * h / w);
      if Result.Y > th then begin
        Result.Y := th;
        Result.X := Trunc(th * w / h);
      end;
    end else begin
      if h < th then th := h;
      Result.Y := th;
      Result.X := Trunc(th * w / h);
      if Result.X > tw then begin
        Result.X := tw;
        Result.Y := Trunc(tw * h / w);
      end;
    end;
  end;
end;

procedure MakeThumbNail(Src, Dst: TBitmap);
var
  x, y, ix, iy, w, h, dx, dy: Integer;
  x1, x2, x3: integer;
  RowDest, RowSource, RowSourceStart: Integer;
  iRatio: Integer;
  Ratio: Single;
  iRed, iGrn, iBlu: Integer;
  pt: PRGB24;
  iSrc, iDst: Integer;
  lutW, lutH: array of Integer;
begin
  if (Src.Width <= Dst.Width) and (Src.Height <= Dst.Height) then begin
    Dst.Assign(Src);
    Exit;
  end;
  w := Dst.Width;
  h := Dst.Height;
  Ratio:= 1 / (w / Src.Width);
  SetLength(lutW, w);
  x1 := 0;
  x2 := Trunc(Ratio);
  for x := 0 to w - 1 do begin
    lutW[x] := x2 - x1;
    x1 := x2;
    x2 := Trunc((x + 2) * Ratio);
  end;
  Ratio:= 1 / (h / Src.Height);
  SetLength(lutH, h);
  x1 := 0;
  x2 := Trunc(Ratio);
  for x := 0 to h - 1 do begin
    lutH[x] := x2 - x1;
    x1 := x2;
    x2 := Trunc((x + 2) * Ratio);
  end;
  RowDest := Integer(Dst.Scanline[0]);
  RowSourceStart := integer(Src.Scanline[0]);
  RowSource := RowSourceStart;
  iDst := ((w * 24 + 31) and not 31) shr 3;
  iSrc := ((Src.Width * 24 + 31) and not 31) shr 3;
  for y := 0 to h - 1 do begin
    dy := lutH[y];
    x1 := 0;
    x3 := 0;
    for x := 0 to w - 1 do begin
      dx := lutW[x];
      iRed := 0;
      iGrn := 0;
      iBlu := 0;
      RowSource := RowSourceStart;
      for iy := 1 to dy do begin
        pt := PRGB24(RowSource + x1);
        for ix := 1 to dx do begin
          iRed := iRed + pt.R;
          iGrn := iGrn + pt.G;
          iBlu := iBlu + pt.B;
          inc(pt);
        end;
        RowSource := RowSource - iSrc;
      end;
      iRatio := $00FFFFFF div (dx * dy);
      pt := PRGB24(RowDest + x3);
      pt.R := (iRed * iRatio) shr 24;
      pt.G := (iGrn * iRatio) shr 24;
      pt.B := (iBlu * iRatio) shr 24;
      x1 := x1 + 3 * dx;
      inc(x3, 3);
    end;
    RowDest := RowDest - iDst;
    RowSourceStart := RowSource;
  end;
end;

// Fast bilinear resampling procedure found at Swiss Delphi Center + my mods...
procedure TfrmMain.appMainDeactivate(Sender: TObject);
begin
  if editView.Visible then editView.Visible:= False;
  ViewMain.IsEditing:= editView.Visible;
  ViewMain.Invalidate;
end;

procedure TfrmMain.BiResample(Src, Dest: TBitmap; Sharpen: Boolean);
type
  PRGB24 = ^TRGB24;
  TRGB24 = record B, G, R: Byte; end;
  PRGBArray = ^TRGBArray;
  TRGBArray = array[0..0] of TRGB24;
var
  x, y, px, py: Integer;
  i, x1, x2, z, z2, iz2: Integer;
  w1, w2, w3, w4: Integer;
  Ratio: Integer;
  sDst, sDstOff: Integer;
  sScanLine: Array of PRGBArray;
  Src1, Src2: PRGBArray;
  C, C1, C2: TRGB24;
  y1, y2, y3, x3, iRed, iGrn, iBlu: Integer;
  p1, p2, p3, p4, p5: PRGB24;
begin
  // ScanLine buffer for Source
  SetLength(sScanLine, Src.Height);
  sDst:= Integer(src.ScanLine[0]);
  sDstOff:= Integer(src.ScanLine[1]) - Integer(sDst);
  for i:= 0 to Src.Height - 1 do begin
    sScanLine[i]:= PRGBArray(sDst);
    sDst:= sDst + sDstOff;
  end;
  // ScanLine for Destiantion
  sDst:= Integer(Dest.ScanLine[0]);
  y1:= sDst;  // only for sharpening...
  sDstOff:= Integer(Dest.ScanLine[1]) - sDst;
  // Ratio is same for width and height
  Ratio:= ((Src.Width - 1) shl 15) div Dest.Width;
  py:= 0;
  for y := 0 to Dest.Height - 1 do
  begin
    Src1:= sScanLine[py shr 15];
    if py shr 15 < Src.Height - 1 then Src2:= sScanLine[(py shr 15) + 1] else Src2:= Src1;
    z2:= py and $7FFF;
    iz2:= $8000 - z2;
    px:= 0;
    for x := 0 to Dest.Width - 1 do
    begin
      x1 := px shr 15;
      x2:= x1 + 1;
      C1:= Src1[x1];
      C2:= Src2[x1];
      z:= px and $7FFF;
      w2:= (z * iz2) shr 15;
      w1:= iz2 - w2;
      w4:= (z * z2) shr 15;
      w3:= z2 - w4;
      C.r:= (C1.R * w1 + Src1[x2].R * w2 + C2.R * w3 + Src2[x2].R * w4) shr 15;
      C.g:= (C1.G * w1 + Src1[x2].G * w2 + C2.G * w3 + Src2[x2].G * w4) shr 15;
      C.b:= (C1.B * w1 + Src2[x2].B * w2 + C2.B * w3 + Src2[x2].B * w4) shr 15;
      // Set destination pixel
      PRGBArray(sDst)[x]:= C;
      Inc(px, Ratio);
    end;
    sDst:= sDst + SDstOff;
    Inc(py, Ratio);
  end;
  if not Sharpen then Exit;
  // Sharpening...
  y2 := y1 + sDstOff;
  y3 := y2 + sDstOff;
  for y := 1 to dest.Height - 2 do begin
    for x := 0 to dest.Width - 3 do begin
      x1 := x * 3;
      x2 := x1 + 3;
      x3 := x1 + 6;
      p1 := PRGB24(y1 + x1);
      p2 := PRGB24(y1 + x3);
      p3 := PRGB24(y2 + x2);
      p4 := PRGB24(y3 + x1);
      p5 := PRGB24(y3 + x3);
      // -15 -11 for less sharpening use -17 - 13
      iRed := (p1.R + p2.R + (p3.R * -15) + p4.R + p5.R) div -11;
      iGrn := (p1.G + p2.G + (p3.G * -15) + p4.G + p5.G) div -11;
      iBlu := (p1.B + p2.B + (p3.B * -15) + p4.B + p5.B) div -11;
      if iRed < 0 then iRed := 0 else if iRed > 255 then iRed := 255;
      if iGrn < 0 then iGrn := 0 else if iGrn > 255 then iGrn := 255;
      if iBlu < 0 then iBlu := 0 else if iBlu > 255 then iBlu := 255;
      pRGB24(y2 + x2).R:= iRed;
      pRGB24(y2 + x2).G:= iGrn;
      pRGB24(y2 + x2).B:= iBlu;
    end;
    inc(y1, sDstOff);
    inc(y2, sDstOff);
    inc(y3, sDstOff);
  end;
  sScanLine:= nil;
end;

// Load a jpeg and make a thumbnail of it...
function TfrmMain.ThumbImage(Filename: string): TBitmap;
var
  s: string;
  tmp: TBitmap;
  ThumbSize: TPoint;
  bmp: TBitmap;
  newW, newH: Integer;
  sf: Integer;
  fail: Boolean;
begin
  Result := nil;
  tmp:= nil;
  bmp:= TBitmap.Create;
  s := Filename;
  if Filename <> '' then begin
    CellJpeg.Scale:= jsFullSize;
    GetJPGSize(Filename, WI, HI);
    if WI < 1 then WI := 1;
    if HI < 1 then HI := 1;
    sf:= Trunc(Min(HI / TH, WI / TW));
    if sf < 0 then sf := 0;
    case sf of
      0..1: CellJpeg.Scale := jsFullSize;
      2..3: CellJpeg.Scale := jsHalf;
      4..7: CellJpeg.Scale := jsQuarter;
      else  CellJpeg.Scale := jsEighth;
    end;
    fail:= False;
    try
      CellJpeg.LoadFromFile(Filename);
    except
      ShowMessage('An error occured when dealing with: ' + Filename);
      fail:= True;
    end;
    if not fail then begin
      bmp.PixelFormat:= pf24Bit;
      bmp.Width:= CellJpeg.Width;
      bmp.Height:= CellJpeg.Height;
      bmp.Canvas.Draw(0, 0, CellJpeg);
      try
        tmp:= TBitmap.Create;
        ThumbSize := CalcThumbSize(bmp.Width, bmp.Height, ThumbSizeW, ThumbSizeH);
        newW := ThumbSize.X;
        newH := ThumbSize.Y;
        if newW <= 0 then newW := 1;
        if newH <= 0 then newH := 1;
        tmp.PixelFormat := pf24Bit;
        tmp.Width := newW;
        tmp.Height := newH;
        MakeThumbNail(bmp, tmp);
      except
        tmp.Free;
        raise;
      end;
      Result := tmp;
    end else Result:= nil;
  end;
  bmp.Free;
end;

function Blend(Color1, Color2: TColor; A: Byte): TColor;
var
  c1, c2: LongInt;
  r, g, b, v1, v2: byte;
begin
  A:= Round(2.55 * A);
  c1 := ColorToRGB(Color1);
  c2 := ColorToRGB(Color2);
  v1:= Byte(c1);
  v2:= Byte(c2);
  r:= A * (v1 - v2) shr 8 + v2;
  v1:= Byte(c1 shr 8);
  v2:= Byte(c2 shr 8);
  g:= A * (v1 - v2) shr 8 + v2;
  v1:= Byte(c1 shr 16);
  v2:= Byte(c2 shr 16);
  b:= A * (v1 - v2) shr 8 + v2;
  Result := (b shl 16) + (g shl 8) + r;
end;

procedure TfrmMain.GenCellColors;
begin
  cHot:= $00FDDE99;
  cgHotStart:= $00FDF5E6;
  cGHotEnd:= $00FDFBF6;
  cSelected:= $00FDCE99;
  cGSelectedStart:= $00FCEFC4;
  cGSelectedEnd:= $00FDF8EF;
  cShadeSelect:= $00F8F3EA;
  cDisabled:= $00D9D9D9;
  cGDisabledStart:= $00EAE9E9;
  cGDisabledEnd:= $00FCFBFB;
  cShadeDisabled:= $00F6F5F5;
  cGHeaderStart:= $00F9F9F9;
  cGHeaderEnd:= $00FEFEFE;
  cGHeaderHotStart:= $00FFEDBD;
  cGHeaderHotEnd:= $00FFF7E3;
  cGHeaderSelStart:= $00FCEABA;
  cGHeaderSelEnd:= $00FCF4E0;
  cBackground:= clWindow;;
  cLineHighLight:= $00FEFBF6;
  CellBkgColor := clWindow;
  CellBrdColor[False, False] := cDisabled;
  CellBrdColor[False, True] := cDisabled;
  CellBrdColor[True, False] := $00B5B5B5;
  CellBrdColor[True, True] := cSelected;
end;

procedure TfrmMain.ItemPaintBasic(Canvas: TCanvas; R: TRect; State: TsvItemState);
var
  C: TColor;
begin
  Canvas.Brush.Style := bsClear;
  if (State = svSelected) or (State = svHot) then begin
    if (viewMain.Focused) and (State = svSelected) then begin
      Canvas.Pen.Color:= cSelected;
      GradientFill(Canvas, R, cGSelectedStart, cGSelectedEnd, tgTopBottom);
    end else
    if (State = svHot) then begin
      Canvas.Pen.Color:= cHot;
      GradientFill(Canvas, R, cGHotStart, cGHotEnd, tgTopBottom);
    end else begin
      Canvas.Pen.Color:= cDisabled;
      GradientFill(Canvas, R, cGDisabledStart, cGDisabledEnd, tgTopBottom);
    end;
    Canvas.Rectangle(R);
    if (viewMain.Focused) then C:= cShadeSelect else C:= cShadeDisabled;
    Canvas.Pen.Color:= C;
    Canvas.MoveTo(R.Left + 1, R.Top + 2);
    Canvas.LineTo(R.Left + 1, R.Bottom - 2);
    Canvas.LineTo(R.Right - 2, R.Bottom - 2);
    Canvas.LineTo(R.Right - 2, R.Top + 1);
    Canvas.Pen.Style:= psSolid;
    Canvas.Pixels[R.Left, R.Top]:= C;
    Canvas.Pixels[R.Left, R.Bottom - 1]:= C;
    Canvas.Pixels[R.Right - 1, R.Top]:= C;
    Canvas.Pixels[R.Right - 1, R.Bottom -1]:= C;
  end;
end;

procedure TfrmMain.SetThumbSize(Value: Integer; UpdateTrackbar: Boolean);
var
  w, h: Integer;
begin
  case Value of
    32..63   : CellJpeg.Scale := jsQuarter;
    64..127  : CellJpeg.Scale := jsHalf;
    128..255 : CellJpeg.Scale := jsFullSize;
    else CellJpeg.Scale := jsEighth;
  end;
  w:= Value;
  h:= Value;
  w := w + 20;
  If CellStyle = -1 then h:= h + 20 else h := h + 40;
  HSX:= (w - 70) shr 1;  // Help for calculating rating...
  viewMain.CellWidth := w;
  viewMain.CellHeight := h;
  CellScale := Value;
  if UpdateTrackbar then begin
    tbSize.OnChange:= nil;
    tbSize.Position:= CellScale;
    tbSize.OnChange:= tbSizeChange;
  end;
  viewMain.CalcView(False);
  if not UpdateTrackBar then viewMain.SetAtTop(-1, vIdx);
end;

procedure TfrmMain.tbSizeChange(Sender: TObject);
begin
  SetThumbSize(tbSize.Position, False);
end;

procedure BiReSample(Src, Dest: TBitmap);
type
  TRGB24 = record B, G, R: Byte; end;
  PRGBArray = ^TRGBArray;
  TRGBArray = array[0..0] of TRGB24;
var
  x, y, px, py: Integer;
  i, x1, x2, z, z2, iz2: Integer;
  w1, w2, w3, w4: Integer;
  Ratio: Integer;
  sDst, sDstOff: Integer;
  sScanLine: Array[0..255] of PRGBArray;  // Obs! Max 255 pixels high
  Src1, Src2: PRGBArray;
  C, C1, C2: TRGB24;
  y1, y2, y3, x3, iRed, iGrn, iBlu: Integer;
  p1, p2, p3, p4, p5: PRGB24;
begin
  // ScanLine buffer for Source
  sDst:= Integer(src.ScanLine[0]);
  sDstOff:= Integer(src.ScanLine[1]) - Integer(sDst);
  for i:= 0 to Src.Height - 1 do begin
    sScanLine[i]:= PRGBArray(sDst);
    sDst:= sDst + sDstOff;
  end;
  // ScanLine for Destiantion
  sDst:= Integer(Dest.ScanLine[0]);
  y1:= sDst;  // only for sharpening...
  sDstOff:= Integer(Dest.ScanLine[1]) - sDst;
  // Ratio is same for width and height
  Ratio:= ((Src.Width - 1) shl 15) div Dest.Width;
  py:= 0;
  for y := 0 to Dest.Height - 1 do
  begin
    Src1:= sScanLine[py shr 15];
    if py shr 15 < Src.Height - 1 then Src2:= sScanLine[py shr 15 + 1] else Src2:= Src1;
    z2:= py and $7FFF;
    iz2:= $8000 - z2;
    px:= 0;
    for x := 0 to Dest.Width - 1 do
    begin
      x1 := px shr 15;
      x2:= x1 + 1;
      C1:= Src1[x1];
      C2:= Src2[x1];
      z:= px and $7FFF;
      w2:= (z * iz2) shr 15;
      w1:= iz2 - w2;
      w4:= (z * z2) shr 15;
      w3:= z2 - w4;
      C.r:= (C1.R * w1 + Src1[x2].R * w2 + C2.R * w3 + Src2[x2].R * w4) shr 15;
      C.g:= (C1.G * w1 + Src1[x2].G * w2 + C2.G * w3 + Src2[x2].G * w4) shr 15;
      C.b:= (C1.B * w1 + Src2[x2].B * w2 + C2.B * w3 + Src2[x2].B * w4) shr 15;
      // Set destination pixel
      PRGBArray(sDst)[x]:= C;
      Inc(px, Ratio);
    end;
    sDst:= sDst + SDstOff;
    Inc(py, Ratio);
  end;
  // Sharpening...
  y2 := y1 + sDstOff;
  y3 := y2 + sDstOff;
  for y := 1 to dest.Height - 2 do
  begin
    for x := 0 to dest.Width - 3 do
    begin
      x1 := x * 3;
      x2 := x1 + 3;
      x3 := x1 + 6;
      p1 := PRGB24(y1 + x1);
      p2 := PRGB24(y1 + x3);
      p3 := PRGB24(y2 + x2);
      p4 := PRGB24(y3 + x1);
      p5 := PRGB24(y3 + x3);
      // -15 -11                       // -17 - 13
      iRed := (p1.R + p2.R + (p3.R * -15) + p4.R + p5.R) div -11;
      iGrn := (p1.G + p2.G + (p3.G * -15) + p4.G + p5.G) div -11;
      iBlu := (p1.B + p2.B + (p3.B * -15) + p4.B + p5.B) div -11;
      if iRed < 0 then iRed := 0 else if iRed > 255 then iRed := 255;
      if iGrn < 0 then iGrn := 0 else if iGrn > 255 then iGrn := 255;
      if iBlu < 0 then iBlu := 0 else if iBlu > 255 then iBlu := 255;
      pRGB24(y2 + x2).R:= iRed;
      pRGB24(y2 + x2).G:= iGrn;
      pRGB24(y2 + x2).B:= iBlu;
    end;
    inc(y1, sDstOff);
    inc(y2, sDstOff);
    inc(y3, sDstOff);
  end;
end;

procedure TfrmMain.PurgeThumbs;
var
  i: Integer;
  p: PCacheItem;

  function OldestImage: Integer;
  var
    i: Integer;
    Oldest: TDateTime;
  begin
    Oldest := Now;
    Result := -1;
    for i := 0 to ThumbsPool.Count - 1 do begin
      p:= ThumbsPool[i];
      if p.Age <= Oldest then begin
        Oldest := p.Age;
        result := i;
      end;
    end;
  end;

begin
  while (PoolSize > MaxPool) and (ThumbsPool.Count > 0) do begin
    i := OldestImage;
    assert(i >= 0);
    p :=  ThumbsPool[i];
    PoolSize:= PoolSize - p.Size;
    p.Bmp.Free;
    Dispose(p);
    ThumbsPool.Delete(i);
  end;
end;

function TfrmMain.ThumbBmp(idx: Integer) : TBitmap;
var
  i: Integer;
  p: PCacheItem;
  t: PThumbData;
  bmp, tmp: TBitmap;
  pt: TPoint;
begin
  Result:= nil;
  // if we have thumbs, see if we can find it...
  if ThumbsPool.Count > 0 then begin
    i:= ThumbsPool.Count - 1;
    while (i >= 0) and (PCacheItem(ThumbsPool[i]).Idx <> Idx) do i:= i - 1;
    if i <> -1 then begin
      p:= ThumbsPool[i];
      if (p.Idx = idx) then begin
        if (p.Scale = CellScale) then begin
          p.Age:= Now;
          Result:= p.Bmp
        end else begin
          PoolSize:= PoolSize - p.Size;
          p.Bmp.Free;
          Dispose(p);
          ThumbsPool.Delete(i);
        end;
      end;
    end;
  end;
  // if we dont have a thumb, make one...
  if Result = nil then begin
    t:= Items[idx];
    if t.Image <> nil then begin
      TMemoryStream(t.Image).Position := 0;
      CellJPEG.LoadFromStream(TMemoryStream(t.Image));
      bmp:= TBitmap.Create;
      bmp.PixelFormat:= pf24Bit;
      pt:= CalcThumbSize(CellJPEG.Width, CellJPEG.Height, CellScale, CellScale);
      if pt.X <> CellJPEG.Width then begin
        tmp:= TBitmap.Create;
        tmp.PixelFormat:= pf24Bit;
        tmp.Width:= CellJPEG.Width;
        tmp.Height:= CellJPEG.Height;
        tmp.Canvas.Draw(0, 0, CellJPEG);
        bmp.Width:= pt.X;
        bmp.Height:= pt.Y;
        if (bmp.Width > 2) and (bmp.Height > 2) then BiResample(tmp, bmp, True) else bmp.Assign(tmp);
        tmp.Free;
      end else begin
        bmp.Width:= CellJPEG.Width;
        bmp.Height:= CellJPEG.Height;
        bmp.Canvas.Draw(0, 0, CellJPEG);
      end;
      New(p);
      p.Idx:= Idx;
      p.Size:= (bmp.Width * bmp.Height) * 3;
      p.Age:= Now;
      p.Scale:= CellScale;
      p.Bmp:= bmp;
      ThumbsPool.Add(p);
      PoolSize:= PoolSize + p.Size;
      Result:= p.Bmp;
      PurgeThumbs;
    end;
  end;
  //Caption:= IntToStr(PoolSize) + ' bytes, count: ' + IntToStr(ThumbsPool.Count);
end;

procedure TfrmMain.viewMainCellHit(Sender: TObject; Canvas: TCanvas; Idx, x,
  y: Integer; var Selected: Boolean);
var
  i: Integer;
begin
  if (y < 20 ) then begin
    i:= 0;
    if (CellStyle = 0) and (y < 20) then i:= 1;
    if i <> CellEditIdx then begin
      CellEditIdx:= i;
      ViewMain.Invalidate;
    end;
  end else begin
    if CellEditIdx <> -1 then ViewMain.Invalidate;
    CellEditIdx:= -1;
  end;
  if CellStyle <> 1 then Exit;
  // Check if over rating stars
  if (x > HSX) and (x < HSX + 70) and (y < 20) then begin
    CellEditIdx:= 2;
    i:= 1 + (x - HSX) div 14;
    if i <> HotStars then begin
      HotStars:= i;
      ViewMain.Invalidate;
    end;
  end else begin
    if HotStars <> -1 then ViewMain.Invalidate;
    HotStars:= -1;
  end;
end;

procedure TfrmMain.viewMainCellPaint(Sender: TObject; Canvas: TCanvas;
  Cell: TRect; IdxA, Idx: Integer; State: TsvItemState);
var
  i, j, x, y: Integer;
  F, S: Boolean;
  R: TRect;
  TW, TH: Integer;
  Txt: string;
  Thumb: PThumbData;
  pt: TPoint;
begin
  Thumb:= PThumbData(Items[idx]);
  pt:= CalcThumbSize(Thumb.ThumbWidth, Thumb.ThumbHeight, CellScale, CellScale);
  TW := pt.X;
  TH := pt.Y;
  ItemPaintBasic(Canvas, Cell, State);
  F := viewMain.Focused;
  S := State = svSelected;
  x := Cell.Left + ((Cell.Right - (Cell.Left + TW)) shr 1);
  y := Cell.Top + ((Cell.Bottom - (Cell.Top + TH)) shr 1);
  y:= y - 10;
  if Thumb.Image <> nil then Canvas.Draw(x, y, ThumbBmp(idx));
  R.Left := X;
  R.Top := Y;
  R.Right := X + TW;
  R.Bottom := Y + TH;
  Canvas.Pen.Color := CellBrdColor[F, S];
  InflateRect(R, 2, 2);
  Canvas.Rectangle(R);
  Canvas.Pen.Color := clWhite;
  InflateRect(R, -1, -1);
  Canvas.Rectangle(R);
  R:= Cell;
  R.Top:= R.Bottom - 20;
  if CellStyle = 0 then begin
    if (CellEditIdx = 1) and (State = svHot) then Canvas.Font.Color:= $00C08000
    else Canvas.Font.Color:= clBlack;
    txt:= Thumb.Name;
    DrawText(Canvas.Handle, PChar(txt), Length(txt), R, DT_END_ELLIPSIS or
      DT_SINGLELINE or DT_NOPREFIX or DT_CENTER or DT_VCENTER);
  end else begin
    R.Left:= Cell.Left + ((Cell.Right - (Cell.Left + 70)) shr 1);
    if (HotStars <> -1) and (State = svHot) then begin
      for i:= 0 to 4 do begin
        if i < HotStars then j:= 1 else j:= 0;
        ilStars.Draw(Canvas, R.Left + (i * 14), R.Top, j, True);
      end;
    end else
      for i:= 0 to 4 do begin
        if i < (Thumb.Rating and $F) then j:= 1 else j:= 0;
        ilStars.Draw(Canvas, R.Left + (i * 14), R.Top, j, True);
      end;
  end;
end;

procedure TfrmMain.ShowStat;
begin
  if Items.Count = 0 then Exit;
  frmStat:= TfrmStat.Create(nil);
  try
    frmStat.ShowModal;
  finally
    frmStat.Free;
  end;
end;

procedure TfrmMain.viewMainKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key = VK_F8 then ShowStat;
end;

procedure TfrmMain.viewMainMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  idx: Integer;
  pt: TPoint;
  p: PThumbData;
  r: TRect;
begin
  vIdx:= ViewMain.ViewIdx;
  idx:= ViewMain.IdxAtXY(Point(x, y), True);
  if idx = - 1 then Exit;
  if (CellEditIdx = 2) then begin   // Stars...
    if (HotStars <> -1) and (idx <> -1) then begin
      p:= Items[idx];
      if p.Rating = HotStars then p.Rating:= 0 else p.Rating:= HotStars;
    end;
  end else if (CellEditIdx = 1) then begin
    EditId:= CellEditIdx;
    EditIdx:= Idx;
    p:= Items[Idx];
    // Needed since IdxAtXY can not be used with GetItemRect ...
    pt.Y:= ViewMain.ItemAtXY(Point(x, y), True);
    if CellEditIdx = 1 then editView.Text:= p.Name;
    r:= ViewMain.GetItemRect(pt.X, pt.Y);
    editView.Left:= R.Left;
    editView.Width:= R.Right - R.Left;
    editView.Top:= ((R.Bottom - 20)) - ViewMain.sbVert.Position;
    editView.Visible:= True;
    editView.SetFocus;
    ViewMain.IsEditing:= True;
  end;
end;

procedure TfrmMain.viewMainMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  vIdx:= viewMain.ViewIdx;
end;

procedure TfrmMain.viewMainResize(Sender: TObject);
begin
  viewMain.SetAtTop(-1, vIdx);
end;

procedure TfrmMain.viewMainSelecting(Sender: TObject; Count: Integer);
begin
  UpdateStatus;
end;

procedure TfrmMain.AddImages(Files: TStringlist);
var
  Entry: PThumbData;
  Dir: string;
  SR: TSearchRec;
  j, p, n: Integer;
  FName: string;
  JPEG: TJPEGImage;
  MS: TMemoryStream;
  ThumbBmp: TBitmap;
  SFI: TSHFileInfo;
  PV: Single;
begin
  frmMain.Enabled := False;
  try
    Forms.Application.ProcessMessages;
    proMain.Position:= 0;
    proMain.Visible:= True;
    PV := 100 / Files.Count;
    JPEG := TJPEGImage.Create;
    JPEG.CompressionQuality := 80;
    TW := frmMain.ThumbSizeW;
    TH := frmMain.ThumbSizeH;
    j := 0;
    viewMain.BeginUpdate;
    while (j < Files.Count) do begin
      FName := Files[j];
      Dir := LowerCase(ExtractFilePath(FName));
      if FindFirst(FName, faAnyFile, SR) = 0 then begin
        New(Entry);
        Entry.Name := ExtractFileName(FName);
        n:= Paths.IndexOf(Dir);
        if (n= -1) then Entry.PathID:= Paths.Add(dir) else Entry.PathID:= n;
        SHGetFileInfo(PChar(FName), FILE_ATTRIBUTE_NORMAL, SFI,
          SizeOf(TSHFileInfo), SHGFI_TYPENAME or SHGFI_SYSICONINDEX);
        Entry.Rating:= 0;
        Entry.Size := SR.Size;
        Entry.Modified := FileDateToDateTime(SR.Time);
        Entry.IWidth := 0;
        Entry.IHeight := 0;
        Entry.ThumbWidth := 0;
        Entry.ThumbHeight := 0;
        ThumbBmp := ThumbImage(FName);
        if ThumbBmp <> nil then begin
          JPEG.Assign(ThumbBmp);
          JPEG.Compress;
          MS:= TMemoryStream.Create;
          try
            JPEG.SaveToStream(MS);
          except
            MS.Free;
            raise;
          end;
          Entry.Image:= MS;
          Entry.ThumbWidth:= ThumbBmp.Width;
          Entry.ThumbHeight:= ThumbBmp.Height;
          ThumbBmp.Free;
          n:= Items.Add(Entry);
          if n <> -1 then viewMain.Items.Add(n);
        end else Dispose(Entry);
      end;
      FindClose(SR);
      p := Trunc(j * PV);
      if p <> proMain.Position then proMain.Position := p;
      Forms.Application.ProcessMessages;
      inc(j);
    end;
    JPEG.Free;
  finally
    frmMain.Enabled := True;
    proMain.Visible:= False;
  end;
  viewMain.EndUpdate;
  tbSizeChange(nil);
  UpdateStatus;
end;

procedure TfrmMain.btnAddClick(Sender: TObject);
var
  Dir: String;
begin
  Dir:= BrowseForFolder('Add images folder', '', False);
  Files.Clear;
  if Dir = '' then Exit;
  FindFiles(Files, Dir);
  AddImages(Files);
end;

procedure TfrmMain.chkRatingClick(Sender: TObject);
begin
  if chkRating.Checked then CellStyle:= 1 else CellStyle:= 0;
  ViewMain.Invalidate;
end;

procedure TfrmMain.editViewExit(Sender: TObject);
begin
  EditView.Visible := False;
  ViewMain.IsEditing:= False;
  ViewMain.SetFocus;
  ViewMain.Invalidate;
end;

procedure TfrmMain.editViewKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
var
  p: PThumbData;
  done: Boolean;
  txt: String;
begin
  if key = VK_Escape then done:= True
  else if key = VK_Return then begin
    done:= True;
    txt:= editView.Text;
    p:= Items[EditIdx];
    if (EditId = 1) then begin
      if p.Name <> txt then begin
        p.Name:= txt;
        ViewMain.Invalidate;
      end;
    end;
  end else done:= False;
  editView.Visible:= not done;
  ViewMain.IsEditing:= editView.Visible;
  if done then ViewMain.SetFocus;
end;

procedure TfrmMain.editViewKeyPress(Sender: TObject; var Key: Char);
begin
  if (Key = #13) or (Key = #27) then Key:= #0;
end;

procedure TfrmMain.ClearThumbsPool;
var
  i: Integer;
  Thumb: PCacheItem;
begin
  for i := ThumbsPool.Count - 1 downto 0 do begin
    Thumb := ThumbsPool[i];
    if Thumb.Bmp <> nil then Thumb.Bmp.Free;
    Dispose(Thumb);
  end;
  ThumbsPool.Clear;
  PoolSize:= 0;
end;

procedure TfrmMain.FormCreate(Sender: TObject);
begin
  Files:= TStringList.Create;
  Paths:= TStringList.Create;
  Items:= TList.Create;
  // Max thumbnail size...
  ThumbSizeW:= 255;
  ThumbSizeH:= 255;
  viewMain.CellWidth := ThumbSizeW + 20;
  viewMain.CellHeight := ThumbSizeH + 40;
  CellJpeg:= TJpegImage.Create;
  GenCellColors;
  CellStyle:= 0;
  PoolSize:= 0;
  MaxPool:= Round(((Screen.Width * Screen.Height) * 3) * 1.5);
  ThumbsPool:= TList.Create;
  VistaGradient(imgTop, 255, 0, clWhite, clBlack, True, True);
  VistaGradient(imgBottom, 192, 0, $00F8F8F8, $00E0E0E0, True, False);
  HotStars:= -1;
  ViewMain.CellSelect:= False;
end;

procedure TfrmMain.FormDestroy(Sender: TObject);
var
  i: Integer;
  Item: PThumbData;
begin
  CellJpeg.Free;
  for i := Items.Count - 1 downto 0 do begin
    Item := Items[i];
    if Item.Size <> 0 then Item.Image.Free;
    Dispose(Item);
  end;
  Items.Clear;
  Items.Free;
  ClearThumbsPool;
  ThumbsPool.Free;
  Paths.Free;
  Files.Free;
end;

procedure TfrmMain.FormResize(Sender: TObject);
begin
  VistaGradient(imgTop, 255, 0, clBtnFace, clSilver, True, True);
  VistaGradient(imgBottom, 192, 0, $00F8F8F8, $00E0E0E0, True, False);
end;

end.
